import * as React from 'react';
import { StyleSheet, TextInput, Platform } from 'react-native';
import { ATEditorProps, ATUser, ATRange, ATListItem } from './types';

const isAndroid = Platform.OS === 'android';
const isIOS = Platform.OS === 'ios';

const DEFAULT_MATCH = /@\[([^\]]+?)\]\(id:([^\]]+?)\)/gim;

const defaultFormatAtUser = (user: ATUser) => `@[${user.username}](id:${user.id})`;

const between = (x: number, min: number, max: number) => x >= min && x <= max;

export default class ATEditor extends React.PureComponent<ATEditorProps> {
  static DEFAULT_MATCH = DEFAULT_MATCH;

  static defaultProps = {
    // autoFocus: true,
    formatAtUser: defaultFormatAtUser,
    matchFormattedTextReg: DEFAULT_MATCH,
  };

  private defaultValue: string = '';
  private value: string = '';
  private atList: ATListItem[] = [];
  private eventCount: number = -1;
  private delaySetSelectionHandle: any = null;
  private textInput: React.RefObject<TextInput> = React.createRef();
  private range!: ATRange;
  private inputText: string = '';
  private selectionPosition?: number;

  constructor(props: ATEditorProps) {
    super(props);
    this.parseFormattedText();
  }

  componentDidMount() {
    if (this.defaultValue) {
      this.triggerChangeCallback(this.defaultValue);
    }
  }

  componentWillUnmount() {
    this.cancelFrame();
  }

  private cancelFrame() {
    this.delaySetSelectionHandle != null && cancelAnimationFrame(this.delaySetSelectionHandle);
  }

  private sortAtList() {
    this.atList.sort((a, b) => a.key[0] - b.key[0]);
  }

  private parseFormattedText() {
    const { defaultValue, matchFormattedTextReg } = this.props;

    if (!defaultValue) {
      return;
    }

    let indexOffset = 0;

    this.defaultValue = defaultValue.replace(
      matchFormattedTextReg!,
      (match: string, username: string, id: string, index: number) => {
        const start = index + indexOffset;
        const usernameLength = username.length;
        const matchLength = match.length;
        this.atList.push({
          key: [start, start + usernameLength + 1],
          user: { username, id },
        });

        // 获取偏移量 @[xxx](id: xxx) -> @xxx
        indexOffset += usernameLength - matchLength + 1;

        return `@${username}`;
      },
    );
    this.value = this.defaultValue;
  }

  private formatWithAtList(inputText: string) {
    const users: ATUser[] = [];
    const len = this.atList.length;

    // 不需要做匹配的时候直接提前返回
    if (!len || !inputText) {
      return {
        users,
        text: inputText,
      };
    }

    const { formatAtUser } = this.props;

    let formattedText: string = '';
    // let formattedNodes: any = [];
    let startIndex = 0;
    let prefix: string;

    for (let i = 0; i < len; i++) {
      const { key: selection, user } = this.atList[i];
      prefix = inputText.slice(startIndex, selection[0]);
      users.push(user);
      formattedText += prefix;
      formattedText += formatAtUser!(user);

      // 维护索引
      startIndex = selection[1];
    }

    // 处理末尾数据
    const suffix = inputText.slice(startIndex);
    if (suffix) {
      formattedText += suffix;
      // formattedNodes.push(suffix);
    }

    return {
      // nodes: formattedNodes,
      users,
      text: formattedText,
    };
  }

  private triggerChangeCallback(newText: string) {
    const { onChange } = this.props;
    // 格式化输入数据
    const { users, text: formattedText } = this.formatWithAtList(newText);

    onChange &&
      onChange({
        displayText: newText,
        text: formattedText,
        users,
      });
  }

  private updateAtList(selection: ATRange, diff: number) {
    for (let i = 0, len = this.atList.length; i < len; i++) {
      const {
        key: [a, b],
      } = this.atList[i];

      if (
        between(a, selection.start, selection.end) ||
        between(b, selection.start, selection.end)
      ) {
        const newKey: [number, number] = [a + diff, b + diff];

        this.atList[i].key = newKey;
      }
    }
  }

  private updateValue(newText: string) {
    const nextProps: any = { text: newText };
    const hasSelectionPosition = this.selectionPosition != null;

    if (isAndroid && hasSelectionPosition) {
      nextProps.selection = {
        start: this.selectionPosition,
        end: this.selectionPosition,
      };
    }

    this.textInput.current!.setNativeProps(nextProps);
    this.value = newText;

    if (isIOS && hasSelectionPosition) {
      this.delaySetSelectionHandle = requestAnimationFrame(() => {
        this.textInput.current!.setNativeProps({
          selection: {
            start: this.selectionPosition,
            end: this.selectionPosition,
          },
        });
        delete this.selectionPosition;
      });
    }
  }

  private getLastNativeSelection() {
    const valueLength = this.value.length;
    return (
      // @ts-ignore
      this.textInput.current._lastNativeSelection || {
        start: valueLength,
        end: valueLength,
      }
    );
  }

  trigger = () => {
    this.onTextInput({
      nativeEvent: {
        text: '@',
        range: this.getLastNativeSelection(),
      },
    });
  };

  addAtUser = (user: ATUser) => {
    const { username } = user;
    const range = this.getLastNativeSelection();
    const text = this.value;
    const usernameLength = username.length;
    const newText = `${text.slice(0, range.start - 1)} @${username} ${text.slice(range.end)}`;
    const selectionStart = range.start;
    const selectionEnd = range.start + usernameLength + 1;

    // 光标位置
    this.selectionPosition = selectionEnd + 1;
    this.updateValue(newText);

    // 数据发生变化后，在光标后面的@的key应该跟着变化
    this.updateAtList(
      {
        start: range.end,
        end: newText.length,
      },
      usernameLength + 2,
    );

    this.atList.push({
      key: [selectionStart, selectionEnd],
      user,
    });

    this.sortAtList();

    this.triggerChangeCallback(newText);
  };

  private onChangeText = (text: string) => {
    const { range, value, inputText } = this;

    let newText: string = `${value.slice(0, range.start)}${inputText}${value.slice(range.end)}`;

    this.updateValue(newText);

    // 数据发生变化后，在光标后面的@的key应该跟着变化
    this.updateAtList(
      {
        start: range.end,
        end: Math.max(value.length, text.length),
      },
      range.start + inputText.length - range.end,
    );

    this.triggerChangeCallback(newText);

    if (inputText === '@') {
      const { onAt } = this.props;
      onAt && onAt();
    }
  };

  private onTextInput = (e: any) => {
    const { text, range, eventCount } = e.nativeEvent;
    // 在选中一段文字然后输入另外文字的时候会触发两次，通过 eventCount 忽略第二次触发
    if (eventCount != null) {
      if (eventCount === this.eventCount) {
        return;
      }
      // 保存本次 eventCount
      this.eventCount = eventCount;
    }
    this.cancelFrame();

    const newRange = { ...range };
    const atList = [...this.atList];

    // 判断选择区域是否在某一个 @用户 的索引范围之内，如果处于索引中间，使其扩充到索引边界
    for (let i = 0, len = this.atList.length; i < len; i++) {
      const { key: selection } = this.atList[i];
      const [a, b] = selection;

      // 判断光标区域的开始是否在某一个 @用户 的索引范围之内
      if (between(newRange.start, a, b)) {
        newRange.start = a;
      }
      // 判断光标区域的结束是否在某一个 @用户 的索引范围之内
      if (between(newRange.end, a, b)) {
        newRange.end = b;
      }
      if (
        // 删除选中区域中所有的 @用户
        between(a, newRange.start, newRange.end) ||
        between(b, newRange.start, newRange.end)
      ) {
        atList.splice(i, 1);
      }
    }
    // console.log(newRange);
    this.atList = atList;
    this.range = newRange;
    this.inputText = text;

    const maybeNewRangeStart = newRange.start + text.length;
    // 判断是否需要维护光标位置
    if (range.start !== newRange.start || range.start === 0 || isAndroid) {
      this.selectionPosition = maybeNewRangeStart;
    } else {
      delete this.selectionPosition;
    }

    if (eventCount == null) {
      this.onChangeText(this.value.slice(0, range.start) + text + this.value.slice(range.end));
    }
  };

  render() {
    const {
      style,
      // scrollEnabled,
      onChange,
      formatAtUser,
      defaultValue,
      matchFormattedTextReg,
      // atTextColor,
      ...restProps
    } = this.props;

    return (
      <TextInput
        {...restProps}
        // @ts-ignore
        onTextInput={this.onTextInput}
        multiline
        ref={this.textInput}
        onChangeText={Platform.select({
          ios: this.onChangeText,
        })}
        defaultValue={this.defaultValue}
        style={[styles.textInput, style]}
        // scrollEnabled={false}
      />
    );
  }
}

const styles = StyleSheet.create({
  textInput: {
    textAlignVertical: 'top',
    flex: 1,
    padding: 10,
    fontSize: 16,
    lineHeight: 20,
  },
});
